from django.contrib import admin
from django.http import HttpResponseRedirect
from django.urls import path
from django.utils.html import format_html

from .models import Rule, RulePath
from .utils import get_ip

# Register your models here.


class RulePathInlineAdmin(admin.TabularInline):
    model = RulePath
    extra = 1


_position_template = """
<div style="display:flex; gap: 8px">
    <a href="{object_id}/rule_start/" title="start">&#x21C8;</a>
    |
    <a href="{object_id}/rule_up/" title="up">&#x2191;</a>
    |
    <a href="{object_id}/rule_down/" title="down">&#x2193;</a>
    |
    <a href="{object_id}/rule_end/" title="end">&#x21CA;</a>
</div>
"""


@admin.register(Rule)
class RuleAdmin(admin.ModelAdmin):
    list_display = ("position_short", "position_buttons", "action", "name", "rule")
    list_display_links = ("position_short",)
    list_editable = ("name", "action", "rule")
    ordering = ("position",)
    fields = ["name", "action", "rule"]
    inlines = [RulePathInlineAdmin]

    @admin.display(description="")
    def position_short(self, obj):
        return obj.position

    @admin.display(ordering="position")
    def position_buttons(self, obj):
        return format_html(_position_template, object_id=obj.id)

    def get_changeform_initial_data(self, request):
        initial = super().get_changeform_initial_data(request)
        initial["rule"] = "*"
        return initial

    def get_urls(self):
        return [
            path("test_rules/", self.test_rules),
            path("test_rules/<path:test_path>", self.test_rules),
            path("<str:object_id>/rule_up/", self.rule_up),
            path("<str:object_id>/rule_down/", self.rule_down),
            path("<str:object_id>/rule_start/", self.rule_start),
            path("<str:object_id>/rule_end/", self.rule_end),
            *super().get_urls(),
        ]

    def rule_up(self, request, object_id):
        self.model.objects.position_up(object_id)
        return HttpResponseRedirect("../../")

    def rule_down(self, request, object_id):
        self.model.objects.position_down(object_id)
        return HttpResponseRedirect("../../")

    def rule_start(self, request, object_id):
        self.model.objects.position_start(object_id)
        return HttpResponseRedirect("../../")

    def rule_end(self, request, object_id):
        self.model.objects.position_end(object_id)
        return HttpResponseRedirect("../../")

    def test_rules(self, request, test_path=None):
        from django.contrib.messages import ERROR, INFO, SUCCESS, WARNING

        test_ip = request.POST.get("test_ip", None)
        if not test_ip:
            test_ip = get_ip(request)

        if test_path is None:
            back_count = 1
            test_path = request.POST.get("test_path", None) or ""
        else:
            back_count = test_path.count("/") + 1
        if test_path:
            rule_id = RulePath.objects.match_path_and_ip(test_path, test_ip)
            self.message_user(
                request,
                f"Parameters: ip: {test_ip}, path: {test_path}",
                level=INFO,
            )
        else:
            rule_id = Rule.objects.match_ip(test_ip)
            self.message_user(
                request,
                f"Parameters: ip: {test_ip}",
                level=INFO,
            )
        if rule_id:
            rule = Rule.objects.get(id=rule_id)
            self.message_user(
                request,
                f"Matched rule: {rule.name}, action: {Rule.ACTION[rule.action]}",
                level=SUCCESS if rule.action == "a" else WARNING,
            )
        else:
            self.message_user(request, "No rule matched", level=ERROR)

        return HttpResponseRedirect("../" * back_count)

    def delete_queryset(self, request, queryset):
        super().delete_queryset(request, queryset)
        Rule.objects.position_cleanup()



from django.conf import settings
from django.http import HttpRequest

from .utils import get_ip


def apply_iprestrict(request: HttpRequest, group):
    from .models import Rule

    rule = Rule.objects.filter(name=group).first()
    if not rule:
        return int(getattr(settings, "IPRESTRICT_DEFAULT_ACTION", "a") == "b")
    if rule.match_ip(get_ip(request), return_action=True) == "b":
        return 1

    return 0



from django.apps import AppConfig
from django.db.models.signals import post_delete


def signal_position_cleanup(**kwargs):
    from .models import Rule

    Rule.objects.position_cleanup()


class DjangoFastIprestrictConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "django_fast_iprestrict"

    def ready(self):
        from .models import Rule

        post_delete.connect(
            signal_position_cleanup,
            sender=Rule,
            dispatch_uid="django-fast-iprestrict-after-deletion",
        )



import ipaddress
import re

from django.core.exceptions import ValidationError


def validate_rule(value):
    if value == "*":
        return
    try:
        ipaddress.ip_network(value, strict=False)
    except ValueError:
        raise ValidationError(
            'Enter a valid IPv4 or IPv6 network or "*".',
            code="invalid",
            params={"value": value},
        )


def validate_regex(value):
    try:
        re.compile(value)
    except re.error:
        raise ValidationError(
            'Invalid regex.',
            code="invalid",
            params={"value": value}
        )



import ipaddress
import re
from contextlib import ExitStack
from functools import lru_cache

from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Max, Window
from django.db.models.functions import RowNumber
from django.db.transaction import atomic

from .validators import validate_regex, validate_rule

# Create your models here.


def get_default_position():
    return Rule.objects.aggregate(Max("position", default=0))["position__max"] + 1


class RuleManager(models.Manager):
    def _atomic_update(self):
        stack = ExitStack()
        stack.queryset = self.all().select_for_update()
        stack.enter_context(atomic())
        return stack

    def _position_cleanup(self, queryset):
        u_dict = {}
        for val in queryset.annotate(
            position_new=Window(expression=RowNumber(), order_by="position"),
        ):
            if val.position != val.position_new:
                u_dict[val.id] = val.position_new
        for rule_id, position in u_dict.items():
            queryset.filter(id=rule_id).update(position=position)
        RulePath.objects.path_ip_matchers.cache_clear()
        self.ip_matchers.cache_clear()

    def position_cleanup(self):
        with self._atomic_update() as stack:
            self._position_cleanup(stack.queryset)

    @lru_cache(maxsize=1)
    def ip_matchers(self):
        patterns = []
        for obj in self.exclude(action="c"):
            patterns.append((obj.get_processed_rule(), obj.id, obj.action))
        return patterns

    def match_ip(self, ip, return_action=False):
        ip_network_user = ipaddress.ip_network(ip, strict=False)
        for ip_network, rule_id, action in self.ip_matchers():
            try:
                if ip_network == "*" or ip_network_user.subnet_of(ip_network):
                    return action if return_action else rule_id
            except TypeError:
                pass
        return None

    def position_start(self, rule_id):
        with self._atomic_update() as stack:
            stack.queryset.filter(id=rule_id).update(position=0)
            self._position_cleanup(stack.queryset)

    def position_end(self, rule_id):
        with self._atomic_update() as stack:
            max_position = stack.queryset.aggregate(
                max_position=Max("position", default=0)
            )["max_position"]
            # ugly hack
            self.get_queryset().annotate().filter(id=rule_id).update(
                position=max_position + 1
            )
            self._position_cleanup(stack.queryset)

    def position_up(self, rule_id):
        with self._atomic_update() as stack:
            rule_position = stack.queryset.filter(id=rule_id).values("position")
            stack.queryset.annotate(
                position_mod=models.Case(
                    models.When(position__lte=1, id=rule_id, then=models.Value(0)),
                    models.When(id=rule_id, then=models.Value(-1)),
                    models.When(
                        position=models.Subquery(rule_position) - models.Value(1),
                        then=models.Value(1),
                    ),
                    default=models.Value(0),
                ),
            ).update(position=models.F("position") + models.F("position_mod"))
            self._position_cleanup(stack.queryset)

    def position_down(self, rule_id):
        with self._atomic_update() as stack:
            rule_position = stack.queryset.filter(id=rule_id).values("position")
            stack.queryset.annotate(
                position_mod=models.Case(
                    models.When(id=rule_id, then=models.Value(1)),
                    models.When(
                        position=models.Subquery(rule_position) + models.Value(1),
                        then=models.Value(-1),
                    ),
                    default=models.Value(0),
                ),
            ).update(position=models.F("position") + models.F("position_mod"))
            self._position_cleanup(stack.queryset)


class Rule(models.Model):
    position = models.PositiveIntegerField(blank=True, default=get_default_position)
    name = models.CharField(max_length=50, unique=True)
    rule = models.CharField(max_length=50, validators=[validate_rule])
    ACTION = {"a": "allow", "b": "deny", "c": "disabled"}
    action = models.CharField(max_length=1, choices=ACTION, default="a")
    objects = RuleManager()

    class Meta:
        ordering = ("position",)

        indexes = [
            models.Index(fields=["position"]),
        ]

    def get_processed_rule(self):
        return (
            ipaddress.ip_network(self.rule, strict=False) if self.rule != "*" else "*"
        )

    def match_ip(self, ip, return_action=False):
        if self.action == "c":
            return None
        ip_network_user = ipaddress.ip_network(ip, strict=False)
        ip_network = self.get_processed_rule()
        try:
            if ip_network == "*" or ip_network_user.subnet_of(ip_network):
                return self.action if return_action else self.id
        except TypeError:
            pass
        return None


class RulePathManager(models.Manager):
    @lru_cache(maxsize=1)
    def path_ip_matchers(self):
        patterns = []
        last_pattern = None
        for obj in self.exclude(rule__action="c").select_related("rule"):
            if not last_pattern or last_pattern[1] != obj.id:
                if last_pattern:
                    patterns.append((re.compile(last_pattern[0]), *last_pattern[1:]))
                last_pattern = [
                    obj.get_processed_path(),
                    obj.rule.get_processed_rule(),
                    obj.rule_id,
                    obj.rule.action,
                ]
            else:
                last_pattern[0] = "%s|%s" % (
                    last_pattern[0],
                    obj.get_processed_path(),
                )
        if last_pattern:
            patterns.append((re.compile(last_pattern[0]), *last_pattern[1:]))
        return patterns

    def match_path_and_ip(self, path, ip, return_action=False):
        ip_network_user = ipaddress.ip_network(ip, strict=False)
        for path_pattern, ip_network, rule_id, action in self.path_ip_matchers():
            try:
                if path_pattern.match(path) and (
                    ip_network == "*" or ip_network_user.subnet_of(ip_network)
                ):
                    return action if return_action else rule_id
            except TypeError:
                pass
        return None


class RulePath(models.Model):
    rule = models.ForeignKey(Rule, related_name="urls", on_delete=models.CASCADE)
    path = models.TextField(max_length=4096)
    is_regex = models.BooleanField(default=False, blank=True)
    objects = RulePathManager()

    class Meta:
        ordering = ("rule__position", "id")

    def get_processed_path(self):
        return self.path if self.is_regex else re.escape(self.path)

    def clean_fields(self, exclude=None):
        super().clean_fields(exclude=exclude)
        if self.is_regex:
            try:
                validate_regex(self.path)
            except ValidationError as exc:
                if exclude and "path" in exclude:
                    raise exc
                else:
                    raise ValidationError({"path": exc})



from asgiref.sync import iscoroutinefunction, sync_to_async
from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.utils.decorators import sync_and_async_middleware

from .utils import get_ip


@sync_and_async_middleware
def fast_iprestrict(get_response):
    from .models import RulePath

    # One-time configuration and initialization goes here.
    if iscoroutinefunction(get_response):
        amatch_path_and_ip = sync_to_async(RulePath.objects.match_path_and_ip)

        async def middleware(request):
            action = await amatch_path_and_ip(
                request.path, get_ip(request), return_action=True
            )
            if not action:
                action = getattr(settings, "IPRESTRICT_DEFAULT_ACTION", "a")
            if action == "b":
                raise PermissionDenied()
            response = await get_response(request)
            return response

    else:

        def middleware(request):
            action = RulePath.objects.match_path_and_i(
                request.path, get_ip(request), return_action=True
            )
            if not action:
                action = getattr(settings, "IPRESTRICT_DEFAULT_ACTION", "a")
            if action == "b":
                raise PermissionDenied()
            response = get_response(request)
            return response

    return middleware
